/* Copyright: GNU GPL v2 */
/* Author: David Muse <dmuse@engr.latech.edu> 1997 */

#include <stdio.h>
#include <string.h>
#include <sys/vfs.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include "surreal.xpm"
#include "hardware.xpm"
#include "purplestoneswirl.xpm"
#include "pixmaps/cd.xpm"
#include "pixmaps/zip.xpm"
#include "pixmaps/harddisk.xpm"
#include "pixmaps/unixfloppy.xpm"
#include "pixmaps/dosfloppy.xpm"
#include "pixmaps/macfloppy.xpm"
#include "pixmaps/nfs.xpm"
#include "pixmaps/dos.xpm"

static Atom  wm_delete_window;
static char font_string1[] = "-*-helvetica-bold-r-*-*-13-240-*-*-*-*-*-*";
static char font_string2[] = "-*-helvetica-medium-r-*-*-13-240-*-*-*-*-*-*";
static int font_string_pos1 = 21;
static int font_string_pos2 = 26;
static int fontsize = 15;
/*
static char font_string1[] = "9x15bold";
static char font_string2[] = "9x15";
static int font_string_pos1 = -1;
static int font_string_pos2 = -1;
static int fontsize = 15;
*/

static int fontheight;
static int popupwidth;
static int popupheight;

struct XpmIcon {
        Pixmap          pixmap;
        Pixmap          mask;
        XpmAttributes   xpmattributes;
};

static int bgwidth, bgheight, mwidth=55, mheight=57, linemaxh=20; 
static int xpos=0, ypos=0;
static int boolpop=0, boolpos=0;
static int depth, hicolor=0;

XtIntervalId timeid=0;

char usercmd[256]="";

XtAppContext    application;
Widget  mainwidget;
Display *display;
int     screen, screenwidth, screenheight;
Window  root, mainwindow, filewindow, mountwindow, mounterwin[24], iconwin[24];
Pixel   fgotherpixel, fgcdpixel, fgnfspixel, 
                fghdxpixel, fghdepixel, fgfdxpixel, fgfdepixel, 
                bgpixel, borderpixel, zeropixel, popupfgpixel, 
                popupbgpixel, popupborderpixel, 
                rootmntpixel, rootumntpixel, usermntpixel, userumntpixel;
GC      mainwindowgc, zerogc, flipgc, thingc, flipthingc, popupgc;
GC      translucentgc;
XGCValues       gcv;
unsigned        long    gcm;
int     interval, poppedup1=0, poppedup2=0;
char    bgcolor[100], zerocolor[100], bgpix[100], 
        fgothercolor[100], fgcdcolor[100], fgnfscolor[100],
        fghdxcolor[100], fghdecolor[100], fgfdxcolor[100], fgfdecolor[100],
        rootmntcolor[100], rootumntcolor[100], 
        usermntcolor[100], userumntcolor[100];
char    bordercolor[100], popupbgpix[100],
        popupbgcolor[100], popupfgcolor[100], popupbordercolor[100];
Pixmap  bgpixmap, canvas;
struct  XpmIcon bgtexture, popupbgpixmap, *mounter[24];
XFontStruct     *popupfont1, *popupfont2;
int     fscounter, fstabcounter;
char    dfcommand[120],
        mountcmd[7] = "mount ",
        umountcmd[8] = "umount ";
int     showzero=0;
struct  XpmIcon harddisk,dos,cd,nfs,unixfloppy,dosfloppy,macfloppy,zip;

struct  fsrecord {
                        char    mountdir[1024];
                        char    filesystem[1024];
                        char    blocks[11];
                        char    used[11];
                        char    free[11];
                        char    percent[8];
                        int     x1, y1, x2, y2;
                        int     mounted;
                        int     user;
                        char    type[12];
                        char    mntopts[1024];
                };

struct  fsrecord        fs[24], fstab[24];

void usage()
{
        printf("\
Usage: xfsm [-int time_interval (seconds)] [-mounts (show mounts)]\n\
            [-cmd user_defined_command] [-hicolor] [-help]\n\
            [-w width] [-h height] [-x xpos] [-y ypos]\n\
            [-maxh line_max_height] [-showzero]\n\
            [-fnsize fontsize] [-fn1 font1] [-fn2 font2]\n\
            [-z zerocolor] [-bg color] [-bp bg_pixmap]\n\
            [-cdrom fg_color_cdrom] [-nfs fg_color_nfs]\n\
            [-hdunix fg_color_hd_ext2] [-hdelse fg_color_hd_else]\n\
            [-fdunix fg_color_fd_ext2] [-fdelse fg_color_fd_else]\n\
            [-popbg popup_bg_color] [-popbp popup_bg_pixmap]\n\
            [-popfg popup_fg_color] [-popbd popup_border_color]\n\
            [-rootm root_mount_color] [-rootu root_umount_color]\n\
            [-userm user_mount_color] [-useru user_umount_color]\n");
        exit(0);
}

Pixel getcolor(char *colorname) {

        XColor  color;
        XWindowAttributes       attributes;
        
        XGetWindowAttributes(display, mainwindow, &attributes);
        color.pixel=0;

        XParseColor(display, attributes.colormap, colorname, &color);
        XAllocColor(display, attributes.colormap, &color);

        return color.pixel;
}

void    defaults1() {
        strcpy(bgcolor, "#c3c3c3\0");
        if (depth<=8) 
           strcpy(bgpix, "None\0");
        else
           strcpy(bgpix, "default\0");
        strcpy(popupbgpix, "None\0");
        strcpy(fgothercolor,  "Cyan\0");
        strcpy(fgcdcolor,  "Magenta\0");
        strcpy(fgnfscolor, "LightSlateGray\0");
        strcpy(fghdxcolor, "DodgerBlue\0");
        strcpy(fghdecolor, "Green\0");
        strcpy(fgfdxcolor, "Orange\0");
        strcpy(fgfdecolor, "Yellow\0");
        strcpy(zerocolor, "IndianRed\0");
        strcpy(rootmntcolor, "LightGreen\0");
        strcpy(rootumntcolor, "Pink\0");
        strcpy(usermntcolor, "Green\0");
        strcpy(userumntcolor, "FireBrick1\0");
        strcpy(popupbgcolor, "Bisque\0");
        strcpy(popupfgcolor, "Black\0");
        strcpy(popupbordercolor, "LightGray\0");
        interval=2;
}

void    defaults2() {
int i, j;
        if (strcasecmp(bgpix,"None")) {
                bgtexture.xpmattributes.valuemask |= (XpmReturnPixels | 
                                                        XpmReturnExtensions );
                                                         
                if (!strcasecmp(bgpix,"default")) {
                  XpmCreatePixmapFromData(display, mainwindow, 
                                        hardware_xpm,
                                        &bgtexture.pixmap, 
                                        &bgtexture.mask, 
                                        &bgtexture.xpmattributes);

                } else {
                        XpmReadFileToPixmap(display, mainwindow, bgpix, 
                                        &bgtexture.pixmap, 
                                        &bgtexture.mask, 
                                        &bgtexture.xpmattributes);
                }
        }

        if (strcasecmp(popupbgpix,"None")) {
                popupbgpixmap.xpmattributes.valuemask |= (XpmReturnPixels | 
                                                        XpmReturnExtensions);
                if (!strcasecmp(popupbgpix,"default")) {
                        XpmCreatePixmapFromData(display, mainwindow, 
                                        purplestoneswirl_xpm,
                                        &popupbgpixmap.pixmap, 
                                        &popupbgpixmap.mask, 
                                        &popupbgpixmap.xpmattributes);
                } else {
                        if (!strcmp(popupbgpix,"Transparent") ||
                                !strcmp(popupbgpix,"Translucent")) {
                                popupbgpixmap.pixmap=0;
                        } else {
                                XpmReadFileToPixmap(display, mainwindow, 
                                        popupbgpix, 
                                        &popupbgpixmap.pixmap, 
                                        &popupbgpixmap.mask, 
                                        &popupbgpixmap.xpmattributes);
                        }
                }
        }
}

void    parsecmdline(int argc, char *argv[]) {

        char    *argument;
        int     i;

        for (i=1; i<argc; i++) {
                argument=argv[i];
                if (argument[0]=='-') {
                  if (!strncmp(argument,"-showzero",9)) {
                           showzero=1;
                  } else
                  if (!strncmp(argument,"-mounts",7)) {
                           boolpop=1;
	          } else
            	  if (!strncmp(argument,"-hicolor",8)) {
                           hicolor=1;
                  } else {
                  if (i<argc-1) {
                        if (!strncmp(argument,"-fn1",4)) {
			   strcpy(font_string1,argv[++i]);
			} else
                        if (!strncmp(argument,"-fn2",4)) {
			   strcpy(font_string2,argv[++i]);
			} else
                        if (!strncmp(argument,"-fnsize",7)) {
                           fontsize = atoi(argv[++i]);
                           if (font_string_pos1>=0) {
                            font_string1[font_string_pos1] = '0'+fontsize/10;
                            font_string1[font_string_pos1+1] = '0'+fontsize%10;
			   }
                           if (font_string_pos2>=0) {
                            font_string2[font_string_pos2] = '0'+fontsize/10;
                            font_string2[font_string_pos2+1] = '0'+fontsize%10;
			   }
                        } else
                        if (!strncmp(argument,"-bg",3)) {
                                strcpy(bgcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-bp",3)) {
                                strcpy(bgpix, argv[++i]);
                        } else
                        if (!strncmp(argument,"-cdrom",6)) {
                                strcpy(fgcdcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-nfs",4)) {
                                strcpy(fgnfscolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-hdunix",7)) {
                                strcpy(fghdxcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-hdelse",7)) {
                                strcpy(fghdecolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-fdunix",7)) {
                                strcpy(fgfdxcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-fdelse",7)) {
                                strcpy(fgfdecolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-bd",3)) {
                                strcpy(bordercolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-popbg",6)) {
                                strcpy(popupbgcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-popbp",6)) {
                                strcpy(popupbgpix, argv[++i]);
                        } else
                        if (!strncmp(argument,"-popfg",6)) {
                                strcpy(popupfgcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-popbd",6)) {
                                strcpy(popupbordercolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-rootm",6)) {
                                strcpy(rootmntcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-rootu",6)) {
                                strcpy(rootumntcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-userm",6)) {
                                strcpy(usermntcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-useru",6)) {
                                strcpy(userumntcolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-z",2)) {
                                strcpy(zerocolor, argv[++i]);
                        } else
                        if (!strncmp(argument,"-int",4)) {
                                interval=atoi(argv[++i]);
                        } else
                        if (!strncmp(argument,"-help",5)) {
                                usage();
                        } else
                        if (!strncmp(argument,"-maxh",5)) {
                                linemaxh=atoi(argv[++i]);
                        } else
                        if (!strncmp(argument,"-w",2)) {
                                mwidth=atoi(argv[++i]);
                        } else
                        if (!strncmp(argument,"-h",2)) {
                                mheight=atoi(argv[++i]);
                        } else
                        if (!strncmp(argument,"-x",2)) {
                                boolpos=1;
                                xpos=atoi(argv[++i]);
                        } else
                        if (!strncmp(argument,"-y",2)) {
                                boolpos=1;
                                ypos=atoi(argv[++i]);
                        } else
                        if (!strncmp(argument,"-cmd",4)) {
                                strcpy(usercmd, argv[++i]);
                        } else usage();
                  } else usage();
		  }
		}
        }
}

void    setuppixmaps() {

        harddisk.xpmattributes.valuemask = (XpmReturnPixels | 
                                                        XpmReturnExtensions);
        cd.xpmattributes.valuemask = (XpmReturnPixels | 
                                                        XpmReturnExtensions);
        zip.xpmattributes.valuemask = (XpmReturnPixels | 
                                                        XpmReturnExtensions);
        dos.xpmattributes.valuemask = (XpmReturnPixels | 
                                                        XpmReturnExtensions);
        unixfloppy.xpmattributes.valuemask = (XpmReturnPixels | 
                                                        XpmReturnExtensions);
        dosfloppy.xpmattributes.valuemask = (XpmReturnPixels | 
                                                        XpmReturnExtensions);
        macfloppy.xpmattributes.valuemask = (XpmReturnPixels | 
                                                        XpmReturnExtensions);
        nfs.xpmattributes.valuemask = (XpmReturnPixels | 
                                                        XpmReturnExtensions);

        XpmCreatePixmapFromData(display, mainwindow, harddisk_xpm,
                                        &harddisk.pixmap, 
                                        &harddisk.mask, 
                                        &harddisk.xpmattributes);
        XpmCreatePixmapFromData(display, mainwindow, dos_xpm,
                                        &dos.pixmap, 
                                        &dos.mask, 
                                        &dos.xpmattributes);
        XpmCreatePixmapFromData(display, mainwindow, cd_xpm,
                                        &cd.pixmap, 
                                        &cd.mask, 
                                        &cd.xpmattributes);
        XpmCreatePixmapFromData(display, mainwindow, nfs_xpm,
                                        &nfs.pixmap, 
                                        &nfs.mask, 
                                        &nfs.xpmattributes);
        XpmCreatePixmapFromData(display, mainwindow, unixfloppy_xpm,
                                        &unixfloppy.pixmap, 
                                        &unixfloppy.mask, 
                                        &unixfloppy.xpmattributes);
        XpmCreatePixmapFromData(display, mainwindow, dosfloppy_xpm,
                                        &dosfloppy.pixmap, 
                                        &dosfloppy.mask, 
                                        &dosfloppy.xpmattributes);
        XpmCreatePixmapFromData(display, mainwindow, macfloppy_xpm,
                                        &macfloppy.pixmap, 
                                        &macfloppy.mask, 
                                        &macfloppy.xpmattributes);
        XpmCreatePixmapFromData(display, mainwindow, zip_xpm,
                                        &zip.pixmap, 
                                        &zip.mask, 
                                        &zip.xpmattributes);

        if (strcasecmp(bgpix,"None")) {
                bgpixmap=bgtexture.pixmap;
                bgwidth=bgtexture.xpmattributes.width;
                bgheight=bgtexture.xpmattributes.height;
        } else {
                bgpixmap=XCreatePixmap(display, mainwindow, mwidth, mheight, 
                                depth);
                bgwidth=mwidth;
                bgheight=mheight;
                XFillRectangle(display, bgpixmap, flipgc, 0, 0, mwidth, mheight);
        }

        canvas=XCreatePixmap(display, mainwindow, mwidth, mheight, depth);

}

void    setupfonts() {
int second;
        popupfont1 = XLoadQueryFont(display, font_string1);
        popupfont2 = XLoadQueryFont(display, font_string2);
        fontheight = popupfont1->ascent + popupfont1->descent + 3;
        second  = popupfont2->ascent + popupfont2->descent + 3;
        if (second>fontheight) fontheight = second;
}

void    readfstabfile() {

        int     i;
        FILE    *fstabfile;
        char    buffer[1024];
        char    filename[12] = "/etc/fstab";

        fstabfile = fopen(filename, "r");

        fstabcounter=0;

        /* initiate buffer to strlen >0 */
        strcpy(buffer,"");
        
        if (fstabfile==NULL) {
                printf("The filesystem description file: %s couldn't be opened.\n",
                        filename);
                exit(0);
        }

        while (feof(fstabfile)==0 && fstabcounter<24) {

                buffer[0]=0;
        
                /* get a line */
                i=0;
                sprintf(buffer,"%c",fgetc(fstabfile));
                while(buffer[i]!=10 && !feof(fstabfile)) {
                        i++;
                        if (!feof(fstabfile)) {
                                sprintf(buffer,"%s%c",
                                        buffer,fgetc(fstabfile));
                        }
                };

                if (!feof(fstabfile)) {
                        if (buffer[0]!='#') {
                                strcpy(fstab[fstabcounter].filesystem,
                                        "");
                                strcpy(fstab[fstabcounter].mountdir,"");
                                strcpy(fstab[fstabcounter].type,"");
                                strcpy(fstab[fstabcounter].mntopts,"");
                                sscanf(buffer,"%s %s %s %s",
                                        fstab[fstabcounter].filesystem,
                                        fstab[fstabcounter].mountdir, 
                                        fstab[fstabcounter].type,
                                        fstab[fstabcounter].mntopts);
                                if (strlen(fstab[fstabcounter].filesystem)>0 &&
                                        strcmp(fstab[fstabcounter].type,"proc") &&
                                        strcmp(fstab[fstabcounter].type,"devpts") &&
                                        strcmp(fstab[fstabcounter].type,"swap") &&
                                        strcmp(fstab[fstabcounter].type,"ignore")) {
                                        fstab[fstabcounter].user=0;
                                        if (strstr(fstab[fstabcounter].mntopts,"user")!=NULL) {
                                                fstab[fstabcounter].user=1;
                                        }
                                        if (!strcmp(fstab[fstabcounter].mountdir,
                                                        "none") ||
                                                !strcmp(fstab[fstabcounter].mountdir, "/") ||
                                                strstr(fstab[fstabcounter].mntopts, "nouser")!=
                                                        NULL) {
                                                fstab[fstabcounter].user=0;
                                        }
                                        fstabcounter++;
                                }
                        }
                }
        }

        fclose(fstabfile);

}

void    readmtabfile()
{
        FILE    *fmtabfile;
        char    buffer[1024];
        char    filename[12] = "/etc/mtab";

        fmtabfile = fopen(filename, "r");

        if (fmtabfile == NULL)
	  {
          fprintf(stderr, "Cannot open or read %s\n", filename);
          return;
	  }

        fscounter = 0;
        while (!feof(fmtabfile))
	  {
          if (fgets(buffer, 1023, fmtabfile))
	    {
            sscanf(buffer, "%s %s %s %s ", fs[fscounter].filesystem, fs[fscounter].mountdir, fs[fscounter].type, fs[fscounter].mntopts);
            if (fs[fscounter].filesystem[0] != '\0' && 
                strcmp(fs[fscounter].mountdir,"none") && 
                strcmp(fs[fscounter].mountdir,"/dev/pts") && 
                strcmp(fs[fscounter].mountdir,"/net") && 
                strcmp(fs[fscounter].mountdir,"/proc")) fscounter++;
	    }
	  }
        fclose(fmtabfile);
}

static long
adjust_blocks (blocks, fromsize, tosize)
     long blocks;
     int fromsize, tosize;
{
  if (tosize <= 0)
    abort ();
  if (fromsize <= 0)
    return -1;
								    
  if (fromsize == tosize)	/* E.g., from 512 to 512.  */
    return blocks;
  else if (fromsize > tosize)	/* E.g., from 2048 to 512.  */
    return blocks * (fromsize / tosize);
  else				/* E.g., from 256 to 512.  */
    return (blocks / (tosize / fromsize));
}

void    getstatfs() {
        struct  statfs fsd;
        int     i, j;
        double  percent;

        readmtabfile();
    
        for (i=0; i<fscounter; i++)
           if( statfs(fs[i].mountdir, &fsd) == 0)
	      {
              sprintf(fs[i].blocks, "%ld", 
                  adjust_blocks(fsd.f_blocks, fsd.f_bsize, 1024));
              sprintf(fs[i].free, "%ld", 
                  adjust_blocks(fsd.f_bavail, fsd.f_bsize, 1024));
              sprintf(fs[i].used, "%ld", 
                  adjust_blocks(fsd.f_blocks-fsd.f_bfree, fsd.f_bsize, 1024));
              if (fsd.f_blocks>0) 
                 {
                 percent = 100.0 *((double)fsd.f_bavail/(double)fsd.f_blocks);
                 sprintf(fs[i].percent, "%.2f %%", percent);
	         }
	      }
}

int     calclineheight(int num) {

        int height;

        if (num>0) {
                height=(mheight-5-num)/num;
                if (height>linemaxh) height=linemaxh;
        } else {
                printf("\nWarning, Line height defaulted to 1.\n");
                height=1;
        }
        return(height);
}

int     calclinewidth(struct fsrecord *fs, int y, int lineheight) {

        float   blocks, used, linewidth;

        blocks=(float)atoi(fs->blocks);
        if (blocks==0.0) {
                linewidth=(mwidth-7);
        } else {
                used=(float)atoi(fs->used);
                linewidth=(int)((used/blocks)*(float)(mwidth-5));
                if (linewidth<2) linewidth=2;
        }

        /* set up fsrecord x,y's */
        fs->x1=0;
        fs->y1=y-(lineheight/2);
        fs->x2=linewidth;
        fs->y2=y+(lineheight/2);

        return linewidth;
}

void    checkusagecb(XtPointer XtP, XtIntervalId *XtI);

void    buildmainwindow() {

        int     i, j, lineheight, y, linewidth;

                /* check mount status */
                for (i=0; i<fstabcounter; i++) mountstatus(i);

                /* restore canvas pixmap */
                for (i=0; i<=mwidth/bgwidth; i++)
                for (j=0; j<=mwidth/bgwidth; j++)
                XCopyArea(display, bgpixmap, canvas, mainwindowgc, 
		   0, 0, mwidth, mheight, i*bgwidth, j*bgheight);
                XtVaSetValues(mainwidget, "backgroundPixmap",bgpixmap,NULL);


                /* calculate line height */
                lineheight=calclineheight(fscounter);

                XSetLineAttributes(display, mainwindowgc, lineheight, 
                                        LineSolid, CapButt, JoinRound);
                XSetLineAttributes(display, zerogc, lineheight, 
                                        LineSolid, CapButt, JoinRound);

                /* draw lines */
                y=1+(lineheight/2);

                for (i=0; i<fscounter; i++) {

                        /* select line foreground color */

                        if (!strcmp(fs[i].type,"ext2")) {
                           if(fs[i].filesystem[5]=='f') 
                                      gcv.foreground=fgfdxpixel; else
                                      gcv.foreground=fghdxpixel;
                        } else
                        if (!strcmp(fs[i].type,"msdos") || 
                            !strcmp(fs[i].type,"vfat")) {
                           if(fs[i].filesystem[5]=='f') 
                                      gcv.foreground=fgfdepixel; 
                           else
                                      gcv.foreground=fghdepixel;
                        } else 
                        if (!strcmp(fs[i].type,"nfs")) {
                                      gcv.foreground=fgnfspixel;
                        } else 
 		        if (!strcmp(fs[i].type,"iso9660")) {
                                      gcv.foreground=fgcdpixel; 
			} else
                                      gcv.foreground=fgotherpixel; 

                        XChangeGC(display, mainwindowgc, GCForeground, &gcv);

                        /* calculate line width */
                        linewidth=calclinewidth(&fs[i], y, lineheight);

                        /* draw new lines */
                        if (atoi(fs[i].blocks)>0) {
                                XDrawLine(display, canvas, mainwindowgc, 
                                        3, y+3, linewidth+3, y+3);
                        } else {
                                XDrawLine(display, canvas, zerogc, 
                                        3, y+3, linewidth+3, y+3);
                        }

                        if (lineheight>2 && linewidth>2) {
                                if (atoi(fs[i].blocks)>0) {
                                        /* raised */
                                        XDrawLine(display, canvas, thingc,
                                                3, y+lineheight/2+3, 
                                                linewidth+3, y+lineheight/2+3);
                                        XDrawLine(display, canvas, flipthingc,
                                                3, y-lineheight/2+3, 
                                                linewidth+3, y-lineheight/2+3);
                                        XDrawLine(display, canvas, flipthingc,
                                                3, y+lineheight/2+3, 
                                                3, y-lineheight/2+3);
                                        XDrawLine(display, canvas, thingc,
                                                linewidth+3, y+lineheight/2+3, 
                                                linewidth+3, y-lineheight/2+3);
                                } else {
                                        /* lowered */
                                        XDrawLine(display, canvas, flipthingc,
                                                3, y+lineheight/2+3, 
                                                linewidth+3, y+lineheight/2+3);
                                        XDrawLine(display, canvas, thingc,
                                                3, y-lineheight/2+3, 
                                                linewidth+3, y-lineheight/2+3);
                                        XDrawLine(display, canvas, thingc,
                                                3, y+lineheight/2+3, 
                                                3, y-lineheight/2+3);
                                        XDrawLine(display, canvas, flipthingc,
                                                linewidth+3, y+lineheight/2+3, 
                                                linewidth+3, y-lineheight/2+3);
                                }
                        }

                        y=y+lineheight+1;
                }

                XCopyArea(display, canvas, mainwindow, mainwindowgc, 
                                0, 0, mwidth, mheight, 0, 0);

                /* get my stuff drawn, harmlessly */
                XPending(display);

                if (timeid) XtRemoveTimeOut(timeid);
                timeid = XtAppAddTimeOut(application, interval*1000, checkusagecb, NULL); 
}

int     mountstatus(int number) {

        int j, retval;
        
        retval = 0;

        for (j=0; j<fscounter; j++) {

                /* if the filesystem is mounted in the right directory... */
	   if (!strcmp(fstab[number].filesystem,fs[j].filesystem)) {
      	         if (!strcmp(fstab[number].mountdir,fs[j].mountdir)) { 
                     strcpy(fs[j].mntopts, fstab[number].mntopts);
                     retval=1;
                     break;
	         }
	      }
	   }

        return retval;
}

void    mounterborder(int number) {
        if (fstab[number].user==1) {
                if (fstab[number].mounted==0) {
                        XSetWindowBorder(display, mounterwin[number], 
                                                userumntpixel);
                } else {
                        XSetWindowBorder(display, mounterwin[number],
                                                usermntpixel);
                }
        } else {
                if (fstab[number].mounted==0) {
                        XSetWindowBorder(display, mounterwin[number],
                                                rootumntpixel);
                } else {
                        XSetWindowBorder(display, mounterwin[number],
                                                rootmntpixel);
                }
        }
}

void    popdownfilewindow() {

        if (poppedup1==1) {
                XDestroyWindow(display, filewindow);
                poppedup1=0;
        }
}

void    checkusage() {

                /* get a list of filesystems */
                getstatfs();
                buildmainwindow();
}

void    checkusagecb(XtPointer XtP, XtIntervalId *XtI) {

        int     i, j, change=0;

                checkusage();
 
                for (i=0; i<fstabcounter ; i++) {
                        j=mountstatus(i);
                        if (j!=fstab[i].mounted) change=1;
                        fstab[i].mounted=j;
                        if (poppedup2==1) mounterborder(i);
                }

                if (poppedup1==1 && change) {
                        popdownfilewindow();
                        buildmainwindow();
                }
}

void    translucent(Window win, int width, int height) {

        int     i;
        static  unsigned char           dotted[2]={ 1, 1 };

        gcm=GCForeground | GCLineStyle | GCCapStyle | 
                                GCLineWidth;
        gcv.foreground=bgpixel;
        gcv.line_style=LineOnOffDash;
        gcv.cap_style=CapButt;
        gcv.line_width=1;
        translucentgc=XCreateGC(display, mainwindow, gcm, &gcv);
        XSetDashes(display, translucentgc, 0, dotted, 2);
        for (i=0; i<height; i++) {
                if (i%2==0) {
                        XDrawLine(display, win, 
                                        translucentgc,
                                        0, i, width, i);
                } else {
                        XDrawLine(display, win, 
                                        translucentgc,
                                        1, i, width, i);
                }
        }
        XFreeGC(display, translucentgc);
}

void    popdownmountwindow() {

        int i;

        if (poppedup2==1) {
                XDestroyWindow(display, mountwindow);
                poppedup2=0;
        }
}

void    popdownmountwindowcb(Widget w, XtPointer unused, 
                        register XButtonPressedEvent *event, 
                        Boolean *continue_to_dispatch) {

        popdownmountwindow();
}

void    popdownfilewindowcb(Widget w, XtPointer unused, 
                        register XButtonPressedEvent *event, 
                        Boolean *continue_to_dispatch) {

        popdownfilewindow();
}


void    popupfilewindow(struct fsrecord fs, int x, int y) {

        XSetWindowAttributes    setwinattr;
        int                     winx, winy, dec_x1, dec_x2;
        unsigned long           pixel;

        if (poppedup1==1) {
             popdownfilewindow();
             buildmainwindow();
        }

        if (poppedup1==0) {

                poppedup1=2;

                if (x>=screenwidth/2) { 
                        winx=x-12-popupwidth;
                } else {
                        winx=x+12;
                }
        
                popupheight = 8 * fontheight + fontheight/5;
                popupwidth  = 14 * fontheight;

                if (y>=screenheight/2) {
                        winy=y-popupheight;
                } else {
                        winy=y;
                }
         
                filewindow=XCreateSimpleWindow(display, root, winx, winy, 
                                                popupwidth, popupheight, 2, 
                                                popupborderpixel, popupbgpixel);
        
                setwinattr.override_redirect=True;
                if (strcasecmp(popupbgpix,"None")) {
                        setwinattr.background_pixmap=popupbgpixmap.pixmap;
                        XChangeWindowAttributes(display, filewindow, 
                                        CWOverrideRedirect | CWBackPixmap, 
                                        &setwinattr);
                } else {
                        XChangeWindowAttributes(display, filewindow, 
                                        CWOverrideRedirect,
                                        &setwinattr);
                }
        
                XSetFont(display, popupgc, popupfont1->fid);
        
                XMapWindow(display, filewindow);
        
                if (!strcmp(popupbgpix,"Translucent")) {
                        translucent(filewindow, popupwidth, popupheight);
                }
               
        
                dec_x1 = fontheight/7;

                XDrawString(display, filewindow, popupgc, dec_x1, fontheight-3, 
                                "File System", 11);
                XDrawString(display, filewindow, popupgc, dec_x1, 2*fontheight-3, 
                                "Mount Dir", 9);
                XDrawString(display, filewindow, popupgc, dec_x1, 3*fontheight-3, 
                                "Mount Type", 10);
                XDrawString(display, filewindow, popupgc, dec_x1, 4*fontheight-3, 
                                "Options", 7);
                XDrawString(display, filewindow, popupgc, dec_x1, 5*fontheight-3, 
                                        "Blocks (kB)", 11);
                XDrawString(display, filewindow, popupgc, dec_x1, 6*fontheight-3, 
                                        "Used (kB)", 9);
                XDrawString(display, filewindow, popupgc, dec_x1, 7*fontheight-3, 
                                        "Free (kB)", 9);
                XDrawString(display, filewindow, popupgc, dec_x1, 8*fontheight-3, 
                                "Percent free", 12);
        
                XSetFont(display, popupgc, popupfont2->fid);
        
                dec_x2 = XTextWidth(popupfont1, "Percent free", 12)+18;

                XDrawString(display, filewindow, popupgc, dec_x2, fontheight-3, 
                                fs.filesystem, strlen(fs.filesystem));
                XDrawString(display, filewindow, popupgc, dec_x2, 2*fontheight-3, 
                                fs.mountdir, strlen(fs.mountdir));
                XDrawString(display, filewindow, popupgc, dec_x2, 3*fontheight-3, 
                                fs.type, strlen(fs.type));
                XDrawString(display, filewindow, popupgc, dec_x2, 4*fontheight-3, 
                                fs.mntopts, strlen(fs.mntopts));
                XDrawString(display, filewindow, popupgc, dec_x2, 5*fontheight-3, 
                                fs.blocks, strlen(fs.blocks));
                XDrawString(display, filewindow, popupgc, dec_x2, 6*fontheight-3, 
                                fs.used, strlen(fs.used));
                XDrawString(display, filewindow, popupgc, dec_x2, 7*fontheight-3, 
                                fs.free, strlen(fs.free));
                XDrawString(display, filewindow, popupgc, dec_x2, 8*fontheight-3, 
                                fs.percent, strlen(fs.percent));
        
                /* get my stuff drawn, harmlessly */
                XPending(display);
        
                poppedup1=1;
        }

}

void    selectxpm(int i) {

        mounter[i]=&harddisk;
        if (!strcmp(fstab[i].type,"RCKRDG") || !strcmp(fstab[i].type,"iso9660") 
                        || !strcmp(fstab[i].type,"ISO9660")
                        || !strcmp(fstab[i].type,"hsfs")) {
                mounter[i]=&cd;
        } 
        if (strstr(fstab[i].filesystem,"zip")!=NULL ||
            strstr(fstab[i].mountdir,"zip")!=NULL) {
                mounter[i]=&zip;
        } 
        if (!strcasecmp(fstab[i].type,"nfs")) {
                mounter[i]=&nfs;
        }
        if (!strcasecmp(fstab[i].type,"dos") || 
              !strcasecmp(fstab[i].type,"msdos")) {
                mounter[i]=&dos;
        }
        if (strstr(fstab[i].filesystem,"fd")!=NULL) {
                mounter[i]=&unixfloppy;
        }
        if (strstr(fstab[i].filesystem,"fd")!=NULL && 
                (!strcmp(fstab[i].type,"DOS") || 
                !strcmp(fstab[i].type,"msdos"))) {
                mounter[i]=&dosfloppy;
        }
        if (strstr(fstab[i].filesystem,"fd")!=NULL && 
	    (!strcmp(fstab[i].type,"hfs"))) {
                mounter[i]=&macfloppy;
        }
}

void    popupmountwindow(int x, int y) {

        XSetWindowAttributes    setwinattr;
        int                     winx, winy, i, longest1, longest2, longest, 
                                winwidth, winheight;
        int                     posx, posy;
        unsigned long           pixel;
        static  unsigned char           dotted[2]={ 1, 1 };

        if (poppedup2==0) {

                poppedup2=2;

                readfstabfile();

                for (i=0; i<fstabcounter; i++) {
                        fstab[i].mounted=mountstatus(i);
                }

                longest=0;
                longest1=0;
                longest2=0;

                /* find longest name */
                for (i=0; i<fstabcounter; i++) {
                        if (strlen(fstab[i].filesystem)>
                                strlen(fstab[longest1].filesystem)) {
                                longest1=i;
                        }
                        if (strlen(fstab[i].mountdir)>
                                strlen(fstab[longest2].filesystem)) {
                                longest2=i;
                        }
                }

                longest1=XTextWidth(popupfont1, fstab[longest1].filesystem, 
                                        strlen(fstab[longest1].filesystem));
                longest2=XTextWidth(popupfont2, fstab[longest2].mountdir,
                                        strlen(fstab[longest2].mountdir));
                longest=longest1;
                if (longest2>longest) { 
                        longest=longest2;
                }

                if (longest<48) {
                        longest=48;
                }

                winwidth=(((fstabcounter-1)/3)+1)*(longest+24);
                winheight=3*(48+2*fontheight+8);

                if (x>=screenwidth/2) { 
                        winx=x-12-winwidth;
                } else {
                        winx=x+12;
                }
        
                if (y>=screenheight/2) {
                        winy=y-winheight;
                } else {
                        winy=y;
                }
         
                mountwindow=XCreateSimpleWindow(display, root, 
                                                winx, winy, 
                                                winwidth, winheight, 2,
                                                popupborderpixel, popupbgpixel);
                XSelectInput(display, mountwindow, ButtonPressMask);

                setwinattr.override_redirect=True;
                if (strcasecmp(popupbgpix,"None")) {
                        setwinattr.background_pixmap=popupbgpixmap.pixmap;
                        XChangeWindowAttributes(display, mountwindow, 
                                        CWOverrideRedirect | CWBackPixmap, 
                                        &setwinattr);
                } else {
                        XChangeWindowAttributes(display, mountwindow, 
                                        CWOverrideRedirect,
                                        &setwinattr);
                }

                XMapWindow(display, mountwindow);
        
                if (!strcmp(popupbgpix,"Translucent")) {
                        translucent(mountwindow, winwidth, winheight);
                }

                for (i=0; i<fstabcounter; i++) {
        
                        XSetFont(display, popupgc, popupfont1->fid);

                        posy=((i%3)*(winheight/3));
                        posx=((i/3)*(longest+24))+3;
        
                        XDrawString(display, mountwindow, popupgc, 
                                posx, posy+fontheight-3, 
                                fstab[i].filesystem, 
                                strlen(fstab[i].filesystem));
        
                        XSetFont(display, popupgc, popupfont2->fid);
        
                        XDrawString(display, mountwindow, popupgc, 
                                posx, posy+2*fontheight-6, 
                                fstab[i].mountdir, strlen(fstab[i].mountdir));
         
                        mounterwin[i]=XCreateSimpleWindow(display, mountwindow,
                                posx, posy+2*fontheight-2, 
                                48, 48, 2, 
                                popupborderpixel, popupbgpixel);

                        mounterborder(i);
                        
                        XSetWindowBackgroundPixmap(display, mounterwin[i], 0);

                        XSelectInput(display, mounterwin[i], ButtonPressMask);
         
                        iconwin[i]=XCreateSimpleWindow(display, 
                                                mounterwin[i], 
                                                0, 0, 48, 48, 0,
                                                popupborderpixel, popupbgpixel);

                        selectxpm(i);
                        
                        XSetWindowBackgroundPixmap(display, iconwin[i], 
                                        mounter[i]->pixmap);

                        XShapeCombineMask(display, iconwin[i], ShapeBounding,
                                        0, 0, mounter[i]->mask, ShapeSet);

                        XMapWindow(display, mounterwin[i]);
                        XMapWindow(display, iconwin[i]);
                }
        
                /* get my stuff drawn, harmlessly */
                XPending(display);
        
                poppedup2=1;
        }

}

void entermainwindowcb(XtPointer XtP, XtIntervalId *XtI)
{
        popdownmountwindow();
        popdownfilewindow();
        buildmainwindow();
}

void    mountunmount(int number) {

        char buf[80];

        if (fstab[number].mounted==1) {
                sprintf(buf, "%s%s",umountcmd,fstab[number].mountdir);
                system(buf);
        } else {
                sprintf(buf, "%s%s",mountcmd,fstab[number].mountdir);
                system(buf);
        }
        checkusage();
        fstab[number].mounted=mountstatus(number);
        mounterborder(number);
}

void    useraction(int number, int button) {
        char buf[80];
        if (strcmp(usercmd,"")!=0) {
                sprintf(buf,"%s %s %s %d %d &", usercmd,
                   fstab[number].filesystem, fstab[number].mountdir,
                   fstab[number].mounted,button);
                system(buf);
        }
}

void    handleclick() {

        Window  dummy;
        int     rootx, rooty, winx, winy, i;
        unsigned        int     dummyui;
        int     lineheight, y, inaline=0;
        unsigned        int     mask;

        checkusage();

        XQueryPointer(display, mainwindow, &dummy, &dummy,
                        &rootx, &rooty, &winx, &winy, &mask);

        lineheight=calclineheight(fscounter);
        y=1+(lineheight/2);

        for (i=0; i<fscounter; i++) {

                calclinewidth(&fs[i], y, lineheight);

                if (winx>=fs[i].x1+3 && winx<=fs[i].x2+3 &&     
                        winy>=fs[i].y1+3 && winy<=fs[i].y2+3) {
                        popupfilewindow(fs[i], rootx, rooty);
                        inaline=1;
                }

                y=y+lineheight+1;
        }

        if (inaline==0 && poppedup2==0) {
                if (boolpos) popupmountwindow(xpos, ypos);
                       else  popupmountwindow(rootx, rooty);
                }
                        
}

void    handleclickcb(Widget w, XtPointer unused, 
                        register XButtonPressedEvent *event, 
                        Boolean *continue_to_dispatch) {

        handleclick();
}

void quit_xfsm()
{
                XFreePixmap(display, harddisk.pixmap);
                XFreePixmap(display, dos.pixmap);
                XFreePixmap(display, cd.pixmap);
                XFreePixmap(display, nfs.pixmap);
                XFreePixmap(display, unixfloppy.pixmap);
                XFreePixmap(display, dosfloppy.pixmap);
                XFreePixmap(display, zip.pixmap);

                XFreePixmap(display, canvas);
                XFreePixmap(display, bgpixmap);
                XFreeGC(display, mainwindowgc);
                XFreeGC(display, flipgc);
                XFreeGC(display, thingc);
                XFreeGC(display, flipthingc);
                XFreeGC(display, popupgc);
                XDestroyWindow(display, mainwindow);
                XCloseDisplay(display);
                exit(0);
}

void    exitprogcb(Widget w, XtPointer unused, 
                        register XButtonPressedEvent *event, 
                        Boolean *continue_to_dispatch) {
        if (event->type==DestroyNotify) 
                quit_xfsm();
}

static XtActionsRec  actions[] = {
/* action_name    routine */
  { "Quit",       quit_xfsm },    /* quit application */
};

main (int argc, char *argv[]) {

        char    *displayname=NULL;
        XEvent  report;
        int     i, exposenum=0, j;
        XTextProperty   name;
        char    *winname="xfsm";

        mainwidget=XtAppInitialize(&application, "xfsm", NULL, 0, &argc, argv, 
                                NULL, NULL, 0);
        XtAppAddActions(application, actions, XtNumber(actions) );

        display=XtDisplay(mainwidget);

        screen=DefaultScreen(display);
        screenwidth=DisplayWidth(display, screen);
        screenheight=DisplayHeight(display, screen);
        depth=DefaultDepth(display, screen);

        defaults1();

        parsecmdline(argc, argv);

        XtResizeWidget(mainwidget, mwidth, mheight, 0);
        XtRealizeWidget(mainwidget);

        if (depth<=8 && hicolor==0) {
           cd_xpm = cd8_xpm;
           dos_xpm = dos8_xpm;
           dosfloppy_xpm = dosfloppy8_xpm;
           harddisk_xpm = harddisk8_xpm;
           nfs_xpm = nfs8_xpm;
           unixfloppy_xpm = unixfloppy8_xpm;
           zip_xpm = zip8_xpm;
	} else {
           cd_xpm = cd256_xpm;
           dos_xpm = dos256_xpm;
           dosfloppy_xpm = dosfloppy256_xpm;
           harddisk_xpm = harddisk256_xpm;
           nfs_xpm = nfs256_xpm;
           unixfloppy_xpm = unixfloppy256_xpm;
           zip_xpm = zip256_xpm;
	}

        root=RootWindow(display, screen);
        mainwindow=XtWindow(mainwidget);
        XtMapWidget(mainwidget);
        defaults2();

        bgpixel=getcolor(bgcolor);
        fgotherpixel=getcolor(fgothercolor);
        fgcdpixel=getcolor(fgcdcolor);
        fghdxpixel=getcolor(fghdxcolor);
        fghdepixel=getcolor(fghdecolor);
        fgfdxpixel=getcolor(fgfdxcolor);
        fgfdepixel=getcolor(fgfdecolor);
        zeropixel=getcolor(zerocolor);
        popupbgpixel=getcolor(popupbgcolor);
        popupfgpixel=getcolor(popupfgcolor);
        popupborderpixel=getcolor(popupbordercolor);
        rootmntpixel=getcolor(rootmntcolor);
        rootumntpixel=getcolor(rootumntcolor);
        usermntpixel=getcolor(usermntcolor);
        userumntpixel=getcolor(userumntcolor);

        gcm=GCForeground | GCBackground | GCGraphicsExposures;
        gcv.foreground=fghdxpixel;
        gcv.background=bgpixel;
        gcv.graphics_exposures=True;
        mainwindowgc=XCreateGC(display, mainwindow, gcm, &gcv);
        gcm=GCForeground;
        gcv.foreground=zeropixel;
        zerogc=XCreateGC(display, mainwindow, gcm, &gcv);
        gcm=GCForeground | GCBackground;
        gcv.background=bgpixel;
        gcv.foreground=bgpixel;
        flipgc=XCreateGC(display, mainwindow, gcm, &gcv);
        gcv.foreground=popupfgpixel;
        gcv.background=popupbgpixel;
        popupgc=XCreateGC(display, mainwindow, gcm, &gcv);
        gcm=GCForeground;
        gcv.foreground=getcolor("#535353");
        thingc=XCreateGC(display, mainwindow, gcm, &gcv);
        XSetLineAttributes(display, thingc, 1, 
                                LineSolid, CapButt, JoinRound);
        gcv.foreground=getcolor("#dbdbdb");
        flipthingc=XCreateGC(display, mainwindow, gcm, &gcv);
        XSetLineAttributes(display, flipthingc, 1, 
                                LineSolid, CapButt, JoinRound);

        readfstabfile();
        setuppixmaps();
        setupfonts();
        if (boolpop) handleclick(); else checkusage();

        XtAddEventHandler(mainwidget, ButtonPressMask, False, 
                    (XtEventHandler) handleclickcb, (XtPointer)0);

        XtAddEventHandler(mainwidget, EnterWindowMask, False, 
                    (XtEventHandler) entermainwindowcb, (XtPointer)0);

        XtAddEventHandler(mainwidget, StructureNotifyMask, False,
                    (XtEventHandler) exitprogcb, (XtPointer)0);

        XtOverrideTranslations(mainwidget,
             XtParseTranslationTable("<Message>WM_PROTOCOLS:Quit()") );
        wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);
        (void) 
        XSetWMProtocols(display, XtWindow(mainwidget), &wm_delete_window, 1);

        XPending(display);

        while (1) {
                
                XtAppNextEvent(application, &report);
                XtDispatchEvent(&report);
                if (poppedup2==1) {
                        for (j=0; j<fstabcounter; j++) {
                                if (report.type==ButtonPress && 
                                        report.xany.display==display &&
                                        report.xany.window==mounterwin[j]) {
                                        if (report.xbutton.button==Button1)
                                                mountunmount(j);
                                        else
                                          useraction(j,report.xbutton.button);
                                }
                        }
                }
        }
}
